using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace skyscraper8.Experimentals.NdsSsu
{
	interface NdsSsuHandler
	{
		void OnNdsSsuError(int pid, NdsSsuError error);
		void OnNdsSsuFileAnnouncement(int pid, ushort tableIdExtension);
		void OnNdsSsuFileComplete(int pid, ushort tableIdExtension, NdsSsuDataMap dataMap);
		void OnNdsSsuProgress(int pid, ushort tableIdExtension, int blocksDone, int blocksTotal, bool theresMore);
		bool TestNdsSsuFileExists(int pid, ushort tableIdExtension);
	}

	public enum NdsSsuError
	{
		WrongTable,
		InvalidMagic,
		SectionSyntaxIndicatorMissing
	}
}
